<?php
/*
	Copyright mydbr.com 2007-2024 https://mydbr.com
	You are free to modify this file
*/


/*
	This will return the latitude and the longitude of a given address
*/

class GoogleMaps {
  static function URLsafe_b64decode($string)
  {
    $data = is_null($string) ? null : str_replace(array('-','_'),array('+','/'),$string);
    $mod4 = strlen($data) % 4;
    if ($mod4) {
      $data .= substr('====', $mod4);
    }
    return base64_decode($data);
  }

  static function URLsafe_b64encode($string)
  {
    $data = base64_encode($string);
    $data = str_replace(array('+','/'),array('-','_'),$data);
    return $data;
  }

  static function getLink($show_link_in_title, $links)
  {
    $url = '';
    $new_window = 'false';
    if (sizeof($links)==1 && !$show_link_in_title) {
      $url = $links[0][1];
      if ($links[0]['target']=='new_window') {
        $new_window = 'true';
      }
    }
    return array($url, $new_window);
  }

  static function ApiKey( $url )
  {
    global $mydbr_defaults;

    if ($mydbr_defaults['google_maps_api_key_use_in_geocode'] && isset($mydbr_defaults['google_maps_api_key']) && $mydbr_defaults['google_maps_api_key'] ) {
      $url .= '&key='.$mydbr_defaults['google_maps_api_key'];
    }
    return $url;
  }

  static function GetCoordinates( $address )
  {
    $lat_long = Extension::geocode($address);
    
    if (is_array($lat_long) && sizeof($lat_long)==2) {
      return $lat_long;
    }
    $lat_long = array();
    $tryit = true;
    $delay = 0;
    $maxtry = 10; /* We'll try the query max 10 times */
    $lat_long[0] = 'NO_RESPONSE_FROM_GOOGLE'; /* In case we go over the limit */
    while($tryit && $maxtry>0) {
      $tryit = false;

      $url = 'https://maps.googleapis.com'.GoogleMaps::ApiKey('/maps/api/geocode/json?address='.urlencode((string) $address).'');

      $error = null;
      $json = get_url_content($url, $error, 20, true);
      if (!$error && $json) {
        $result = json_decode((string) $json);
        if ($result->status == 'OK') {
          // OK, we got the location
          $lat_long[0] = $result->results[0]->geometry->location->lat;
          $lat_long[1] = $result->results[0]->geometry->location->lng;
        } elseif ($result->status == 'OVER_QUERY_LIMIT') {
          $tryit = true;
          $maxtry--;
          $delay += 100000;
          usleep($delay);
        } else {
          // There was an error, let's pass that to user
          $lat_long[0] = $result->status;
        }
      } else {
        if ($error) {
          $lat_long[0] = $error;
        }
        if ($json) {
          $lat_long[0] = $json;
        }

        // If the file_get_contents failed we will not get a result, we'll just try again (for few times)
        // Don't bother trying if Google said "403"
        if (!strpos($error, "403")) {
          usleep($delay);
          $maxtry--;
          $tryit = true;
        }
      }
    }
    if (sizeof($lat_long)==2) {
      Extension::geocode_set($address, $lat_long);
    }
    return $lat_long;
  }
}

/*
	This is the function described in extensions.php
	$id: object id in the report
	$options: command options described in extensions.php cmds, contains the parsed values from the report
	$dataIn: the 2-dim data array in
	$colInfo: column info - not needed in Google Maps
*/
function Ext_GoogleMaps( $id, $options,  $dataIn, $colInfo )
{
  $output = '';
  // Default size of the map
  $defWidth = 800;
  $defHeight = 600;
  // Zeros will cause automatic centering using markers
  $autoMapX = 0;
  $autoMapY = 0;
  // Zero will cause automatic zooming around markers
  $autoZoom = 0;

  $heatmap = false;
  $heatmap_options = '';

  if (isset($options['dbr.googlemaps.heatmap']['enabled'])) {
    $heatmap = true;
    if (isset($options['dbr.googlemaps.heatmap']['options'])) {
      $heatmap_options = ($heatmap_options ? "," : '').$options['dbr.googlemaps.heatmap']['options'];
    }
  }

  $kml = isset($options['dbr.googlemaps.kml']['url']) ? $options['dbr.googlemaps.kml']['url'] : null;
  if ($kml && !is_array($kml)) {
    $kml = array($kml);
  }
  $geojson = isset($options['dbr.googlemaps.geojson']['url']) ? $options['dbr.googlemaps.geojson']['url'] : [];
  if ($geojson) {
    if (!is_array($options['dbr.googlemaps.geojson']['url'])) {
      $options['dbr.googlemaps.geojson']['url'] = array($options['dbr.googlemaps.geojson']['url']);
    }
    if (!isset($options['dbr.googlemaps.geojson']['style'])) {
      $options['dbr.googlemaps.geojson']['style'] = [];
    }
    if (!is_array($options['dbr.googlemaps.geojson']['style'])) {
      $options['dbr.googlemaps.geojson']['style'] = array($options['dbr.googlemaps.geojson']['style']);
    }
    $geojson = $options['dbr.googlemaps.geojson'];
  }
  
  $polyline = isset($options['dbr.googlemaps.polyline']) ? $options['dbr.googlemaps.polyline'] : null;
  $route = isset($options['dbr.googlemaps.route']) ? $options['dbr.googlemaps.route'] : null;
  $show_link_in_title = false;

  if (isset($options['dbr.googlemaps.link_in_title']['enabled'])) {
    $show_link_in_title = $options['dbr.googlemaps.link_in_title']['enabled'];
  }
  // dbr.googlemaps extension has 2 possible modes: coordinates and address. Lets make sure report has defined one.
  $mode = isset($options['dbr.googlemaps']['mode']) ? $options['dbr.googlemaps']['mode'] : '';

  if (!in_array($mode, array('coordinates', 'address', 'coordinates_address'))) {
    error_print("Usage: select 'dbr.googlemaps', 'coordinates | address | coordinates_address' {,title, width, height, x, y, zoom}");
    return;
  }

  // Does the report define map dimensions?
  $width = isset($options['dbr.googlemaps']['width']) ? $options['dbr.googlemaps']['width'] : $defWidth;
  $height = isset($options['dbr.googlemaps']['height']) ? $options['dbr.googlemaps']['height'] : $defHeight;
  if (is_numeric(substr($width,-1))) {
    $width .= 'px';
  }
  if (is_numeric(substr($height,-1))) {
    $height .= 'px';
  }

  // Does the report define map center location?
  $x = isset($options['dbr.googlemaps']['lat']) ? floatval($options['dbr.googlemaps']['lat']) : $autoMapX;
  $y = isset($options['dbr.googlemaps']['lng']) ? floatval($options['dbr.googlemaps']['lng']) : $autoMapY;
  $zoom = isset($options['dbr.googlemaps']['zoom']) ? intval($options['dbr.googlemaps']['zoom']) : $autoZoom;
  $map_type = isset($options['dbr.googlemaps']['map_type']) ? $options['dbr.googlemaps']['map_type'] : 'roadmap';

  // Automatic map center?
  $noXY = ($x==0 && $y==0) ? 'true' : 'false';
  // Automatic zoom?
  $noZoom = ($zoom == 0) ?  'true' : 'false';
  $output .= '<div '.Extension::resultclass(true, 'map_wrapper', true).'>';
  // If the title attribute has been set we'll use subtitle class to draw it
  if (isset($options['dbr.googlemaps']['title'])) {
    $output .= '<div class="subtitle" style="text-align: center;padding-bottom:2px;">'. $options['dbr.googlemaps']['title'] . '</div>';
  }
  $output .= '<div id="map'.$id.'" class="mydbr_googlemap" style="display:flex;justify-content: center;align-items: center;margin-left:auto;font-size:20px;margin-right:auto;border:thin solid #bbb; width:'.$width.'; height:'.$height.'">Generating the map...</div>';
  $output .= '</div>';
  $output .= "<script>
  if (!window.hasOwnProperty('Google_maps') || Google_maps === undefined) { var Google_maps = []; }
  ";
  
  $output .= "Google_map.init_map('$id',$x,$y,$zoom,'$map_type',".($route ? 'true' : 'false').");";
  
  if ($kml) {
    foreach ($kml as $kml_url) {
      $output .= "Google_map.loadKML('$kml_url');";
    }
  }
  if ($geojson) {
    for ($i=0; $i < sizeof($geojson['url']); $i++) { 
      $geojson_url = $geojson['url'][$i];
      $geojson_style = '';
      if (isset($geojson['style'][$i])) {
        $geojson_style = $geojson['style'][$i];
      }
      $output .= "Google_map.loadGeoJSON('$geojson_url','$geojson_style');";
    }
  }
  $notFound = array();
  // Go through the data to set the markers (pins)
  $heatmap_points = '';
  $heatmap_locations = '';
  $heatmap_weighted = false;
  for ($i=0; $i<sizeof($dataIn) ; $i++) {
    $links = Extension::get_report_links($dataIn[$i], null, true);
    $icon = '';
    $iconShadow = '';
    $markerLabel = '';
    $ok = true;
    $bubble = '';
    if ($i==0 && sizeof($dataIn[0])>2) {
      $heatmap_weighted = true;
    }
    switch ($mode) {
      case 'address':
        $coord = GoogleMaps::GetCoordinates($dataIn[$i][0]);
        if (sizeof($coord)==2) {
          // Got the address, place the marker
          $lat = $coord[0];
          $long = $coord[1];
          if ($heatmap) {
            if ($heatmap_weighted) $weight = $dataIn[$i][1];
          } else {
            if (isset($dataIn[$i][1])) {
              $bubble = $dataIn[$i][1];
              if (sizeof($dataIn[$i])>2) {
                $icon = $dataIn[$i][2];
                if (sizeof($dataIn[$i])>3) {
                  $iconShadow = $dataIn[$i][3];
                  if (sizeof($dataIn[$i])>4) {
                    $markerLabel = $dataIn[$i][4];
                  }
                }
              }
            } else {
              if (!$route) {
                $bubble = $dataIn[$i][0];
              } else {
                $bubble = 'blue';
              }
            }
          }
        } else {
          $ok = false;
          $google_errorCode = '';
          if (sizeof($coord)>0) {
            $google_errorCode = ', '.$coord[0];
          }
          if ($dataIn[$i][0]) {
            array_push($notFound, $dataIn[$i][0].$google_errorCode);
          } else {
            array_push($notFound, 'Missing address for '.isset($dataIn[$i][1]) ? $dataIn[$i][1] : '');
          }
        }
        break;
      case 'coordinates':
        // In this case we already have the coordinates, just place the markers
        $lat = $dataIn[$i][0];
        $long = $dataIn[$i][1];
        if ($heatmap) {
          if ($heatmap_weighted) $weight = $dataIn[$i][2];
        } else {
          if (isset($dataIn[$i][2])) {
            $bubble = $dataIn[$i][2];
            if (sizeof($dataIn[$i])>3) {
              $icon = $dataIn[$i][3];
              if (sizeof($dataIn[$i])>4) {
                $iconShadow = $dataIn[$i][4];
                if (sizeof($dataIn[$i])>5) {
                  $markerLabel = $dataIn[$i][5];
                }
              }
            }
          }
        }
        break;
      case 'coordinates_address':
        $ok = false;
        // In this case we already have the coordinates, just place the markers
        if (!is_null($dataIn[$i][0])) {
          $lat = $dataIn[$i][0];
          $long = $dataIn[$i][1];
          $ok = true;
        } else {
          $coord = GoogleMaps::GetCoordinates($dataIn[$i][2]);
          if (sizeof($coord)==2) {
            $lat = $coord[0];
            $long = $coord[1];
            $ok = true;
          }
        }
        if ($ok) {
          if ($heatmap) {
            if ($heatmap_weighted) $weight = $dataIn[$i][3];
          } else {
            if (isset($dataIn[$i][3])) {
              $bubble = $dataIn[$i][3];
              if (sizeof($dataIn[$i])>4) {
                $icon = $dataIn[$i][4];
                if (sizeof($dataIn[$i])>5) {
                  $iconShadow = $dataIn[$i][5];
                  if (sizeof($dataIn[$i])>6) {
                    $markerLabel = $dataIn[$i][6];
                  }
                }
              }
            }
          }
        }
        break;
    }
    if ($ok) {
      if ($heatmap) {
        if ($heatmap_weighted) {
          $weight = floatval($weight);
        } else {
          $weight = 'null';
        }
        $output .= "Google_map.heatmap_point_add($lat, $long, $weight);";
        
        $comma = (($i==(sizeof($dataIn)-1)) ? '' : ',');
        $heatmap_points .= $comma;
        if ($heatmap_weighted) {
          $heatmap_locations .= $comma;
        }
      } else {
        $link_url = '';
        foreach ($links as $key => $link) {
          $link_url .= '<br><a href="'.$link[1].'">'.htmlentities($link[0]).'</a>';
        }
        list($url, $new_window) = GoogleMaps::getLink($show_link_in_title, $links);
        $output .= "Google_map.new_marker($lat, $long, ".json_encode($bubble.$link_url).",".json_encode($url).", $new_window, $noXY, $noZoom, '$icon', '$iconShadow',".json_encode($markerLabel).");";
      }
    }
  }

  if ($heatmap) {
    $output .= "Google_map.heatmap_options_add('".$heatmap_options."');";
  }
  if ($polyline) {
    $c = $polyline['color'];
    $w = $polyline['width'];
    $output .= "Google_map.polyline('$c',$w);";
  }
  if ($route) {
    $hideMarker = isset($route['hidemarker']) ? ($route['hidemarker'] ? 'true' : 'false') : 'false';
    $output .= "Google_map.route('".$route['travelmode']."',".$hideMarker.');';
  }
  
  if (isset($options['dbr.googlemaps.marker'])) {
    $mm = $options['dbr.googlemaps.marker'];
    for ($i=0; $i < sizeof($mm['lat']); $i++) { 
      if (!isset($mm['lng'][$i]) || !$mm['lng'][$i]) {
        $coord = $coord = GoogleMaps::GetCoordinates($dataIn[$i][0]);
        if (sizeof($lat_lng)==2) {
          $lat = $coord[0];
          $long = $coord[1];
        } else {
          continue;
        }
      } else {
        $lat = $mm['lat'][$i];
        $long = $mm['lng'][$i];
        $bubble = isset($mm['bubble'][$i]) ? $mm['bubble'][$i] : '';
        $icon = isset($mm['icon'][$i]) ? $mm['icon'][$i] : '';
        $iconShadow = isset($mm['iconShadow'][$i]) ? $mm['iconShadow'][$i] : '';
        $markerLabel = isset($mm['markerLabel'][$i]) ? $mm['markerLabel'][$i] : '';
        $output .= "Google_map.new_marker_extra($lat, $long, ".json_encode($bubble).",'', '', '', '', '".$icon."', '$iconShadow',".json_encode($markerLabel).');';
      }
      
    }
  }
  
  $output .= "Google_map.set_bounds($noXY, $noZoom);";
  $output .= "if (!$noZoom) { Google_map.fix_zoom($zoom); }";

  $output .= 'Google_maps.push(Object.assign({}, Google_map));</script>';
  if (sizeof($notFound)>0) {
    $output .= '<div class="comment">Google Maps was not able to locate following addresses:<br>';
    foreach ($notFound as $missing) {
      $output .= '- '.$missing.'<br>';
    }
    $output .= '</div>';
  }
  echo $output;
}
