<?php

abstract class PushSender {
  public $errors = 0;
  public $error_text;
  public $options = array();
  public $app_api_token;
  public $valid_options;
  public $private_key;

  /**
   * Creates a new sender based on hash passed as parameters
   *
   */
  static function create( $params, $options ) {
    $result = null;
    if (is_null($params['sender'])) return null;
    switch (strtolower($params['sender'])) {
      case 'pushbullet':
        $result = new PushbulletSender( $params['pushbullet_access_token'], $options );
        break;
      case 'pushover':
        $result = new PushoverSender( $params['pushover_app_api_token'], $options );
        break;
      case 'prowl':
        $result = new ProwlSender( $options );
        break;
      case 'pushsafer':
        $key = isset($params['pushsafer_private_key']) ? $params['pushsafer_private_key'] : null;
        $result = new PushsaferSender( $key, $options );
        break;
      default:
        # code...
        break;
    }
    return $result;
  }
  /**
   * Delivers a message to a recipient.
   *
   * @param string $recipient   Recipient key
   * @param string $title       Message title
   * @param string $message     Message to be delivered
   * @return boolean            true if message is send, false if errors occurred, Error message is stored in $errors
   */
  abstract function deliver( $id, $recipient, $title, $message );
  
  function get_options( $cmd, $options, $has_alias = false )
  {
    $r = array();
    if (isset($options[$cmd])) {
      $cmd_options = $options[$cmd];
      $size = is_array($cmd_options['option']) ? sizeof($cmd_options['option']) : 1;
      for ($i=0; $i < $size; $i++) { 
        $option = is_array($cmd_options['option']) ? $cmd_options['option'][$i] : $cmd_options['option'];
        $value = is_array($cmd_options['value']) ? $cmd_options['value'][$i] : $cmd_options['value'];
        if ($has_alias) {
          $key = array_search($option, $this->valid_options);
          if ($key !== false) {
            $r[$key] = $value;
          }
        } else {
          if (in_array($option, $this->valid_options)) {
            $r[$option] = $value;
          }
        }
      }
    }
    $this->options = $r;
  }
}

class PushbulletSender extends PushSender {
  
  function __construct( $access_token, $options )	{
    $this->errors = 0;
    $this->error_text = array();
    $this->app_api_token = $access_token;
    $this->valid_options = array('access_token', 'target');
    
    $this->get_options('dbr.push.option', $options);
  }
  
  /**
   * Delivers a message
   *
   * @param string $recipient 	In international format without + in the beginning
   * @param string $message 		Message
   * @return boolean				    whether delivery is successful
   */
  function deliver( $id, $recipient, $title, $message ) {

    if (isset($this->options['access_token'])) {
      $this->app_api_token = $this->options['access_token'];
    }
    $target = isset($this->options['target']) ? $this->options['target'] : 'email';

    // device_iden - Send the push to a specific device. Appears as target_device_iden on the push. You can find this using the /v2/devices call.
    // email - Send the push to this email address. If that email address is associated with a Pushbullet user, we will send it directly to that user, otherwise we will fallback to sending an email to the email address (this will also happen if a user exists but has no devices registered).
    // channel_tag - Send the push to all subscribers to your channel that has this tag.
    // client_iden - Send the push to all users who have granted access to your OAuth client that has this iden.
    
    if (!in_array($target, array('device_iden', 'email', 'channel_tag', 'client_iden'))) {
      // We need to report this just once
      if ($id==0) {
        $this->errors += 1;
        $this->error_text[$id][0] = 'Invalid target for Pushbullet';
      }
      return false;
    }
    
    $postfields = array(
      $target => $recipient, 
      "type" => 'note', 
      "title" => $title,
      "body" => $message
    );

    $ch = curl_init();
    curl_setopt($ch, CURLOPT_URL,            "https://api.pushbullet.com/v2/pushes" );
    curl_setopt($ch, CURLOPT_RETURNTRANSFER, 1 );
    curl_setopt($ch, CURLOPT_POST,           1 );
    curl_setopt($ch, CURLOPT_USERPWD,        $this->app_api_token.":");
    curl_setopt($ch, CURLOPT_POSTFIELDS,     json_encode($postfields) );
    curl_setopt($ch, CURLOPT_HTTPHEADER,     array('Content-Type: application/json')); 
    Extension::curl_proxy($ch);

    $result = curl_exec($ch);
    if ($result===false) {
      $this->errors += 1;
      $this->error_text[$id][0] = curl_error($ch);
    } else {
      $result = json_decode((string) $result);
      if (isset($result->error->message)) {
        $this->errors += 1;
        $this->error_text[$id] = $result->error->message;
      }
    }
    curl_close($ch);
    return true;
  }
}


class ProwlSender extends PushSender {
  
  function __construct( $options )	{
    $this->errors = 0;
    $this->error_text = array();
    $this->valid_options = array('providerkey', 'priority', 'application', 'url');
    
    $this->get_options('dbr.push.option', $options);
  }
  
  /**
   * Delivers a message
   *
   * @param string $recipient 	In international format without + in the beginning
   * @param string $message 		Message
   * @return boolean				    whether delivery is successful
   */
  function deliver( $id, $recipient, $title, $message ) {
    
    $postfields = array(
      'apikey'      =>     $recipient,
      'priority'    =>     0,
      'application' =>     'myDBR',
      'event'       =>     $title,
      'description' =>     $message
    );
    
    foreach ($this->options as $key => $value) {
      $postfields[$key] = $value;
    }
    
    $ch = curl_init('https://api.prowlapp.com/publicapi/add' );
    curl_setopt($ch, CURLOPT_HEADER, 0);
    curl_setopt($ch, CURLOPT_USERAGENT, "myDBR");
    curl_setopt($ch, CURLOPT_RETURNTRANSFER, 1);
    curl_setopt($ch, CURLOPT_CONNECTTIMEOUT,10);
    curl_setopt($ch, CURLOPT_POST, 1);
    curl_setopt($ch, CURLOPT_POSTFIELDS, $postfields);
    Extension::curl_proxy($ch);
    
    $result = curl_exec($ch);
    curl_close($ch);
    
    if ($result === false) {
      $this->error_text[$id][0] = curl_error($ch);
    } else {
      $result = new SimpleXMLElement($result);

      if (!isset($result->success)) {
        $this->errors += 1;
        if (isset($result->error['code'])) {
          switch($result->error['code']) {
            case 400: $this->error_text[$id][0] = 'Bad request, the parameters you provided did not validate.'; break;
            case 401: $this->error_text[$id][0] = 'Not authorized, the API key given is not valid, and does not correspond to a user.'; break;
            case 406: $this->error_text[$id][0] = 'Not acceptable, your IP address has exceeded the API limit.'; break;
            case 409: $this->error_text[$id][0] = 'Not approved, the user has yet to approve your retrieve request'; break;
            case 500: $this->error_text[$id][0] = 'Internal server error, something failed to execute properly on the Prowl side.'; break;
            default:  $this->error_text[$id][0] = 'Unkown error ('.$result->error['code'].')'; break;
          }
        } else {
          $this->error_text[$id][0] = 'Unkown error. No error code';
        }
        return false;
      }
    }
    return true;
  }
}

class PushsaferSender extends PushSender {
  
  function __construct( $private_key, $options )	{
    $this->errors = 0;
    $this->error_text = array();
    $this->private_key = $private_key;
    
    $this->valid_options = array(
      'k' => 'private_key',
      'd' => 'device', 
      's' => 'sound', 
      'v' => 'vibration',
      'i' => 'icon',
      'c' => 'iconcolor',
      'u' => 'url',
      'ut' => 'urltitle',
      'p' => 'picture',
      'is' => 'imagesize',
      'l' => 'timetolive',
      'pr' => 'priority',
      're' => 'retry',
      'ex' => 'expire',
      'a' => 'answer',
      'ao' => 'answeroptions',
      'af' => 'forceanswer',
      'cr' => 'confirm',
      'g' => 'giphygifcode',
    );
    
    $this->get_options('dbr.push.option', $options, true);
  }
  
  function verify_option($key, $value) {
    if (in_array($key, ['d', 'g', 'ao', 'k'])) {
      return ['status' => 'ok', 'value' => $value];
    }
    if (in_array($key, ['u', 'ut'])) {
      return ['status' => 'ok', 'value' => urlencode($value)];
    }
    $ranges = [
      's' => ['min' => 0, 'max' => 63],
      'v' => ['min' => 1, 'max' => 3],
      'i' => ['min' => 1, 'max' => 181],
      'is' => ['min' => 0, 'max' => 3],
      'pr' => ['min' => -2, 'max' => 2],
      're' => ['min' => 60, 'max' => 10800],
      'ex' => ['min' => 60, 'max' => 10800],
      'cr' => ['min' => 60, 'max' => 10800],
      'a' => ['min' => 0, 'max' => 1],
      'af' => ['min' => 0, 'max' => 1],
    ];
    
    if (isset($ranges[$key])) {
      $value = intval($value);
      if ($value >= $ranges[$key]['min'] && $value <= $ranges[$key]['max']) {
        return ['status' => 'ok', 'value' => $value];
      } else {
        return ['status' => 'error', 'error' => 'Unknown '.$key.' option value '.$value];
      }
    }
    return ['status' => 'error', 'error' => 'Unknown option value '.$key];
  }

  
  /**
   * Delivers a message
   *
   * @param string $recipient 	In international format without + in the beginning
   * @param string $message 		Message
   * @return boolean				    whether delivery is successful
   */
  function deliver( $id, $recipient, $title, $message ) {
    // All devices by default
    if (!isset($this->options['d'])) {
      $this->options['d'] = 'a';
    }
    foreach ($this->options as $key => $value) {
      $v = $this->verify_option($key, $value);
      if ($v['status']=='ok') {
        $postfields[$key] = $v['value'];
      } else {
        $this->errors += 1;
        $this->error_text[$id][0] = $v['error'];
        return false;
      }
    }
    if (!$this->private_key && !isset($this->options['k'])) {
      $this->errors += 1;
      $this->error_text[$id][0] = 'No private key defined in $mydbr_push or passed in as option "private_key"';
      return false;
    }
    
    $postfields = [
      'k' =>     $this->private_key ? $this->private_key : $this->options['k'],
      't' =>     urlencode($title),
      'm' =>     urlencode($message)
    ];
    
    $postString = http_build_query($postfields, '', '&');
    $ch = curl_init('https://www.pushsafer.com/api' );
    curl_setopt($ch, CURLOPT_POST, 1);
    curl_setopt($ch, CURLOPT_POSTFIELDS, $postString);
    curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
    curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, false );
    Extension::curl_proxy($ch);
    
    $result = curl_exec($ch);
    if ($result === false) {
      $this->errors += 1;
      $this->error_text[$id][0] = curl_error($ch);
    } else {
      $result = json_decode((string) $result);
      if ($result->status==0) {
        $this->errors += 1;
        $this->error_text[$id][0] = $result->error;
      }
    }
    curl_close($ch);
    return true;
  }
}


class PushoverSender extends PushSender {
  
  function __construct( $app_api_token, $options )	{
    $this->errors = 0;
    $this->error_text = array();
    $this->app_api_token = $app_api_token;
    $this->valid_options = array('token', 'device', 'url', 'url_title', 'priority', 'timestamp', 'sound', 'html');
    
    $this->get_options('dbr.push.option', $options);
  }
  
  /**
   * Delivers a message
   *
   * @param string $recipient 	In international format without + in the beginning
   * @param string $message 		Message
   * @return boolean				    whether delivery is successful
   */
  function deliver( $id, $recipient, $title, $message ) {
    $postfields = array(
      "token" => $this->app_api_token,
      "user" => $recipient, 
      "title" => $title,
      "message" => $message,
    );
    foreach ($this->options as $key => $value) {
      $postfields[$key] = $value;
    }
    $ch = curl_init();
    curl_setopt_array(
      $ch, array(
        CURLOPT_URL => "https://api.pushover.net/1/messages.json",
        CURLOPT_RETURNTRANSFER => true,
        CURLOPT_POSTFIELDS => $postfields
      )
    );
    Extension::curl_proxy($ch);
    
    $result = curl_exec($ch);
    curl_close($ch);
    
    if ($result===false) {
      $this->errors += 1;
      $this->error_text[$id][0] = curl_error($ch);
      return false;
    } else {
      $result = json_decode((string) $result);
      if ($result->status!=1) {
        $this->errors += 1;
        $this->error_text[$id] = $result->errors;
        return false;
      }
    }
    return true;
  }
}


/**
 * Sends Push Messages.
 *
 * @param string $id
 * @param array $options
 * @param array $dataIn
 * @param array $colInfo
 * @return void
 */
function Ext_Push( $id, $options,  $dataIn, $colInfo )
{
  global $mydbr_push;
  
  $push_extension = new Extension;
  
  if (isset($options['dbr.push.sender']['sender'])) {
    $mydbr_push['sender'] = $options['dbr.push.sender']['sender'];
  }
  
  if (!isset($mydbr_push['sender'])) {
    echo '<div class="error">Define the $mydbr_push in user/extension_init.php.</div>';
    return;
  }

  $sender = PushSender::create( $mydbr_push, $options );
  
  if ( $sender == null ) {
    echo '<div class="error">Push sender ' . $mydbr_push['sender'] . ' cannot be used.</div>';
    return;
  }
  $no_sends = 0;
  if (sizeof($dataIn)>0 && sizeof($dataIn[0])>2) {
    for ($i=0; $i<sizeof($dataIn) ; $i++) {
      $sent = $sender->deliver( $i, $dataIn[$i][0], $dataIn[$i][1], $dataIn[$i][2] );
      if (!$sent) {
        $no_sends++;
      }
    }
  } else {
    $sender->errors++;
    $sender->error_text[0][0] = 'Push requires recipient, title and message as parameters';
  }

  $notify_successful_push = true;
  if (isset($options['dbr.push.notify_successful_push']['noyes'])) {
    $notify_successful_push = $options['dbr.push.notify_successful_push']['noyes'];
  }
  if ($sender->errors>0) {
    $txt = "Sent ".(sizeof($dataIn)-$no_sends) . " messages with ".$sender->errors." errors";
    if ($sender->errors==1 && sizeof($sender->error_text) == 1) {
      $av = array_values($sender->error_text);
      $e = array_shift($av);
      $txt .= '.<br>Error: '. (is_array($e) ? $e[0] : $e); 
    }
  } else {
    $txt = "Sent ".sizeof( $dataIn) . (sizeof($dataIn)>1 ? " messages" : " message");
  }
  
  if ($notify_successful_push) {
    echo '<div class="comment">'. $txt . '</div>';
  }
  
  if (isset($options['dbr.push.log.proc']['procedure'])) {
    $msg = '';
    for ($i=0; $i<sizeof($dataIn) ; $i++) {
      $msg .= 'Recipient: '.$dataIn[$i][0].', Status: '.(isset($sender->error_text[$i][0]) ? $sender->error_text[$i][0] : 'OK')."\n";
    }
    $push_extension->push_log( $options['dbr.push.log.proc']['procedure'], $txt, $msg );
  }
  
  if (isset($options['dbr.push.debug']['debug']) && $options['dbr.push.debug']['debug']) {
    echo '<div class="box debug_code"><pre><b>Debug</b><br><br>';
    
    for ($i=0; $i<sizeof($dataIn) ; $i++) {
      echo 'ID: '. ($i+1) . ', Recipient: '.$dataIn[$i][0].', Status: '.(isset($sender->error_text[$i][0]) ? $sender->error_text[$i][0] : 'OK')."\n";
    }
    echo '</pre></div>';
  }
}
