<?php

abstract class SMSSender {

  public $errors;
  public $simulate;


  /**
   * Creates a new sender based on hash passed as parameters
   *
   */
  static function create( $params ) {
    $result = null;
    if (is_array($params) && isset($params['sender']) && isset($params['apiKey']) && $params['apiKey'] ) {
      switch ($params['sender']) {
        case 'clickatell':
          $result = new ClickatellSender( $params['apiKey'] );
          break;
        case 'textlocalin':
          $result = new TextlocalInSender( $params['apiKey'] );
          break;
      }
    }
    return $result;
  }


  /**
   * Delivers a message to a recipient.
   *
   * @param string $recipient 	Telephone number for the recipient, format according to SMS gateway
   * @param string $message 		Message to be delivered
   * @param string $from 			Optional "from" for message, might not be supported by all gateways
   * @return boolean				true if message is send, false if errors occurred, Error message is stored in $errors
   */
  abstract function deliver( $recipient, $message, $from = null );


}

class ClickatellSender extends SMSSender {

  protected $base_url = 'https://platform.clickatell.com/messages/http';
  protected $url;

  /**
   * Creates url for SMS
   *
   * @param string $apiKey
   */
  function __construct( $apiKey )	{
    $this->base_url = $this->base_url . '/send?apiKey=' .$apiKey;
  }

  /**
   * Delivers a message
   *
   * @param string $recipient 	In international format without + in the beginning
   * @param string $message 		Message
   * @param string $from 			Sender ID, must be requested from Clickatell
   * @return boolean				weather delivery is successful
   */
  function deliver( $recipient, $message, $from = null ) {
    // convert to latin-1
    $message = mb_convert_encoding( $message, 'ISO-8859-1', mb_detect_encoding( $message) );

    $url = $this->base_url . "&to=$recipient&content=" . urlencode((string)  $message );
    if ( $from )
      $url .= '&from=' . urlencode((string)  $from );

    if ( $this->simulate ) {
      echo "<p>" . $url .  "</p>";
      $sent[0] = '{"messages":[{"apiMessageId":"123456789","accepted":true,"to":"'.$recipient.'","errorCode":null,"error":null,"errorDescription":null}],"errorCode":null,"error":null,"errorDescription":null}';
    } else {
      $sent = file( $url );
    }
    try {
      $sent = json_decode((string) $sent[0]);
      if (is_object($sent)) {
        if (isset($sent->messages) && is_array($sent->messages) && sizeof($sent->messages)>0) {
          $sent = $sent->messages[0];
          if (is_null($sent->error)) {
            return true;
          } else {
            throw new Exception("Send message failed: ".$sent->error.' '.$sent->errorDescription);
          }
        }
      } else {
        throw new Exception('Send message failed');
      }
    } catch (Exception $e) {
      $this->errors = $e->getMessage();
      return false;
    }
    return true;
  }
}

class TextlocalInSender extends SMSSender {

  protected $base_url = 'https://api.textlocal.in/send/';
  protected $url;
  protected $apiKey;

  /**
   * Creates url for SMS
   *
   * @param string $apiKey
   */
  function __construct( $apiKey )	{
    $this->apiKey = $apiKey;
  }

  /**
   * Delivers a message
   *
   * @param string $recipient 	In international format without + in the beginning
   * @param string $message 		Message
   * @param string $from 			Sender ID, must be requested from Clickatell
   * @return boolean				weather delivery is successful
   */
  function deliver( $recipient, $message, $from = null ) {
    
    if (is_null($from)) {
      $from = 'TXTLCL';
    }
    $apiKey = urlencode((string) $this->apiKey);

    // Send the POST request with cURL
    $ch = curl_init($this->base_url);
    curl_setopt($ch, CURLOPT_POST, true);
    curl_setopt($ch, CURLOPT_POSTFIELDS, 
      array(
        'apikey' => $apiKey, 
        'numbers' => $recipient, 
        'sender' => urlencode((string) $from), 
        'message' => rawurlencode( (string) $message)
      )
    );
    curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
    $response = curl_exec($ch);
    
    $httpCode = curl_getinfo($ch, CURLINFO_HTTP_CODE);
    $error = curl_error($ch);
    curl_close($ch);
    
    if ($response === false) {
      $this->errors = 'Failed to connect to the Textlocal service: ' . $error;
      return false;
    } elseif ($httpCode != 200) {
      $this->errors = 'Bad response from the Textlocal service: HTTP code ' . $httpCode;
      return false;
    }
    $result = json_decode((string) $response);
    if (!is_null($result->errors)) {
      $this->errors  = $result->errors[0]->message;
      return false;
    }
    return true;
  }
}

/**
 * Sends SMS Messages.
 *
 * @param string $id
 * @param array $options
 * @param array $dataIn
 * @param array $colInfo
 * @return void
 */
function Ext_SMS( $id, $options,  $dataIn, $colInfo )
{
  global $mydbr_sms;

  if (!function_exists('mb_convert_encoding')) {
    echo '<div class="error">SMS extension requires that the <a href="http://www.php.net/manual/en/book.mbstring.php">Multibyte String</a> PHP support is installed.</div>';
    return;
  }

  $sender = SMSSender::create( $mydbr_sms );
  if ( $sender == null ) {
    echo '<div class="error">You seem to be missing the Clicatell apiKey definition. See <a href="https://mydbr.com/doc/?extensions.sms.html">SMS Extension documentation</a> for configuration help.</div>';
    return;
  }

  if ( isset($options['dbr.sms']['simulate'] ) )
    $sender->simulate = true;

  if ( isset($options['dbr.sms.sender']['from']))
    $from = $options['dbr.sms.sender']['from'];
  else
    $from = null;

  for ($i=0; $i<sizeof($dataIn) ; $i++) {
    if ( $sender->errors == null )
      $sender->deliver( $dataIn[$i][0], $dataIn[$i][1], $from );
  }

  if ($sender->errors != null) {
    $txt = $sender->errors;
  } else {
    $txt = $sender->simulate ? "Simulated " : "Sent ";
    $txt .= sizeof( $dataIn) . " messages";
  }
  echo '<div class="comment">'. $txt . '</div>';
}
